(*  Title:      Tools/isac/Knowledge/Build_Thydata.thy
    Author:     Walther Neuper, JKU Linz
    (c) due to copyright terms

Collect all knowledge defined for the isac mathengine
*)

theory Build_Thydata
imports
  Isac_Knowledge
  Test_Build_Thydata (*for test/../build_thydata.sml, thy_hierarchy.sml*)
begin

subsection \<open>Build <Theory>-Data\<close>
text \<open>
  <Theory>-Data represent Isac's "deductive knowledge" (whereas <Problems> and <Methods> 
  represent application-oriented and algorithmic knowledge respectively.) <Theory>-Data 
  comprise theories from ProgLang and from Knowledge.

  Note that under normal circumstances, theories are only available from the given formal
  context, not the global theory database (\<^ML_structure>\<open>Thy_Info\<close>). The following
  operations always work:

    \<^item> \<^ML>\<open>@{theory Pure}\<close>, \<^ML>\<open>@{theory Main}\<close>, \<^ML>\<open>@{theory Complex_Main}\<close>
    \<^item> \<^ML>\<open>Theory.ancestors_of\<close>, or better \<^ML>\<open>Theory.nodes_of\<close>
    \<^item> \<^ML>\<open>Context.get_theory\<close> for situations of explicit name-lookup (short or long)

  In contrast, old-style \<^ML>\<open>Thy_Info.get_theory\<close> requires a finished logic image, for
  example in \<^file>\<open>$ISABELLE_ISAC_POS/Tools/isac/Test_Some.thy\<close> loaded from session Isac.

  All theories below ML_structure\<open>MathEngine\<close> in the theories' dependency graph
  can be considered as Isac's version of Pure: the respective code handles items
  declared later between \<^ML>\<open>@{theory Base_Tools}\<close> and \<^ML>\<open>@{theory Isac_Knowledge}\<close>.
  These items are contained in the  ML_structure\<open>Context\<close> in current Isabelle,
  thus all ML_structure\<open>Know_Store\<close> can be replaced by more native Isabelle mechanisms --
  with two exceptions
    \<^item> \<^ML>\<open>Know_Store.add_pbls\<close> and \<^ML>\<open>Know_Store.get_pbls\<close>, 
      handling a tree of \<^ML_type>\<open>Problem.T\<close>
    \<^item> \<^ML>\<open>Know_Store.add_mets\<close> and \<^ML>\<open>Know_Store.get_mets\<close>, 
      handling a tree of \<^ML_type>\<open>MethodC.T\<close>
  where both trees have a structure independent from the dependency graph 
  between \<^ML>\<open>@{theory Base_Tools}\<close> and \<^ML>\<open>@{theory Isac_Knowledge}\<close>.

\<close>

section \<open>Link Know_Store to completed IsacKnowledge\<close>
ML \<open>Know_Store.set_ref_last_thy @{theory}\<close>

ML \<open>
\<close> ML \<open>
\<close> ML \<open>
\<close> 

end
